import * as THREE from '../build/three.module.js';

import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';

import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { GLTFLoader } from './jsm/loaders/GLTFLoader.js';
import { EffectComposer } from './jsm/postprocessing/EffectComposer.js';
import { RenderPass } from './jsm/postprocessing/RenderPass.js';
import { UnrealBloomPass } from './jsm/postprocessing/UnrealBloomPass.js';

class App {
    constructor() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        //renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.setPixelRatio(window.devicePixelRatio);
        //renderer.shadowMap.enabled =  true;
        renderer.toneMapping = THREE.ReinhardToneMapping;
        domWebGL.appendChild(renderer.domElement);

        this._renderer = renderer;

        const scene = new THREE.Scene();
        this._scene = scene;

        this._clock = new THREE.Clock();

        this._setupCamera();
        this._setupLight();
        this._setupPostprocessing();
        //this._setupInteractions();
        this._setupModel();
        this._setupGUI();
        this._setupControls();
        //this._setupHelper();

        window.onresize = this.resize.bind(this);

        this.resize();
        this.render()
    }

    _setupInteractions() {}

    _setupPostprocessing() {
        const renderScene = new RenderPass(this._scene, this._camera);

        const params = {
            exposure: 1,
            bloomThreshold: 0.85,
            bloomStrength: 1.5,
            bloomRadius: 0.4
        };

        const bloomPass = new UnrealBloomPass(new THREE.Vector2(window.innerWidth, window.innerHeight), 1.5, 0.4, 0.85);
        bloomPass.threshold = params.bloomThreshold;
        bloomPass.strength = params.bloomStrength;
        bloomPass.radius = params.bloomRadius;

        const composer = new EffectComposer(this._renderer);
        composer.addPass(renderScene);
        composer.addPass(bloomPass);

        this._composer = composer;
        this._params = params;
        this._bloomPass = bloomPass;
    }

    _setupModel() {
        new GLTFLoader().load("models/gltf/PrimaryIonDrive.glb", (gltf) => {
            const model = gltf.scene;
            this._scene.add(model);

            const mixer = new THREE.AnimationMixer(model);
            const clip = gltf.animations[0];
            mixer.clipAction(clip.optimize()).play();
            this._mixer = mixer;
        });
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;

        const gui = new GUI();
        gui.add(this._params, "exposure", 0.1, 2).onChange((value) => {
            this._renderer.toneMappingExposure = Math.pow(value, 4.0);
        });

        gui.add(this._params, "bloomThreshold", 0.0, 1.0).onChange((value) => {
            this._bloomPass.threshold = Number(value);
        });

        gui.add(this._params, "bloomStrength", 0.0, 3.0).onChange((value) => {
            this._bloomPass.strength = Number(value);
        });

        gui.add(this._params, "bloomRadius", 0.0, 1.0).step(0.01).onChange((value) => {
            this._bloomPass.radius = Number(value);
        });
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(40, window.innerWidth / window.innerHeight, 1, 100);
        camera.position.set(-5, 2.5, -3.5);
        this._scene.add(camera);
        this._camera = camera;
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this._renderer.domElement);
        controls.maxPolarAngle = Math.PI * 0.5;
        controls.minDistance = 1;
        controls.maxDistance = 10;
    }

    _setupLight() {
        const ambientLight = new THREE.AmbientLight(0x404040);
        this._scene.add(ambientLight);

        const pointLight = new THREE.PointLight(0xffffff, 1);
        this._camera.add(pointLight);
        
    }

    update() {
        const delta = this._clock.getDelta();
        this._mixer.update(delta);

        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        if(this._mixer) {
            this.update();
            //this._renderer.render(this._scene, this._camera);
            this._composer.render();
        }
    }

    resize() {
        const width = window.innerWidth;
        const height = window.innerHeight;

        this._camera.aspect = width / height;
        this._camera.updateProjectionMatrix();
        
        this._renderer.setSize(width, height);
        this._composer.setSize(width, height);
    }
}

window.onload = function () {
    new App();
}